/***********************************************
* Jacobi Solver with OpenACC
* ----------------------------------------------
* Authors: Sandra Wienke, RWTH Aachen University
*          Jiri Kraus,    NVIDIA Corporation
************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <omp.h>
#include <openacc.h>

static void initGPU(int argc, char** argv);

int main(int argc, char** argv)
{
    int i,j;
    int n = 4096;
    int m = 4096;
    int iter_max = 100;
    int n_cpu;
    if (argc > 2) n_cpu = atoi(argv[2]);
    else n_cpu = 4000;
    int j_cpu_start = n - n_cpu;
    if (n_cpu < 1 || n_cpu > n-1) {printf("Given number of rows on CPU is not valid.\n");exit(-1);}

    const double pi = 2.0 * asin(1.0);
    const double tol = 1.0e-5;
    double err = 1.0;
    double acc_err = 1.0;

    double* restrict const A	= malloc( n*m * sizeof(double) );
    double* restrict const Anew	= malloc( n*m * sizeof(double) );
    double* restrict const y0	= malloc( n   * sizeof(double) );

#ifdef _OPENACC
  initGPU(argc, argv);
#endif

    // ensure memory locality on cc-NUMA with first touch
    #pragma omp parallel for default(none) shared(A,Anew,m,n,j_cpu_start) private(i,j)
    for( j = j_cpu_start-1; j < n; j++)
    {
        for( i = 0; i < m; i++ )
        {
            Anew[j *m+ i] 	= 0.0;
            A[j *m+ i] 		= 0.0;
        }
    }

    // set boundary conditions
    #pragma omp parallel for default(none) shared(A,m,n) private(i)
    for (i = 0; i < m; i++)
    {
    	A[0	    *m+ i] = 0.0;
    	A[(n-1) *m+ i] = 0.0;
    }

    #pragma omp parallel for default(none) shared(A,Anew,y0,m,n) private(j)
    for (j = 0; j < n; j++)
    {
        y0[j] = sin(pi * j / (n-1));
        A[j *m+ 0] = y0[j];
        A[j *m+ (m-1)] = y0[j]*exp(-pi);
        Anew[j *m+ 0] = A[j *m+ 0];
        Anew[j *m+ (m-1)] = A[j *m+ (m-1)];
    }

  // print in parallel region to get number of threads
  #pragma omp parallel
  {
    #pragma omp master
    {
      printf("Jacobi relaxation Calculation: %d x %d mesh with %d threads. %d rows out of %d on CPU.\n", n, m,omp_get_num_threads(),n_cpu,n);
    }
  }
    double starttime = omp_get_wtime();
    int iter = 0;

  //TODO: adapt data range to copy on GPU corresponding to matrix decomposition
  //TODO: create acc_err explicitly on the device
  #pragma acc data copy(A[0:n*m]) create(Anew[0:n*m])
  {
    while ( err > tol && iter < iter_max ) {

        acc_err = 0.0;
        err = 0.0;

        //TODO: update explicitly acc_err on GPU to defer the copy back of the reduction variable for true heterogeneous computing

        // do one part on GPU
        //TODO: adapt the present data on the GPU corresponding to the matrix decomposition and add the reduction clause
        //TODO: Enable asynchronous computations
        #pragma acc parallel loop present(Anew[0:n*m],A[0:n*m]) reduction(max:acc_err)
        for( j = 1; j < j_cpu_start; j++) {
            for( i = 1; i < m-1; i++ ) {
                Anew[j *m+ i] = 0.25 * ( A[j     *m+ (i+1)] + A[j     *m+ (i-1)]
                                     +   A[(j-1) *m+ i]     + A[(j+1) *m+ i]);
                acc_err = fmax(acc_err,fabs(Anew[j*m+i]-A[j*m+i]));
            }
        }

        // do other part on CPU
        //TODO: To fully utilize the CPU use OpenMP to execute this loop in parallel

        for( j = j_cpu_start; j<n-1; j++) {
            for( i = 1; i < m-1; i++ ) {
                Anew[j *m+ i] = 0.25 * ( A[j     *m+ (i+1)] + A[j     *m+ (i-1)]
                                     +   A[(j-1) *m+ i]     + A[(j+1) *m+ i]);
                err = fmax(err,fabs(Anew[j*m+i]-A[j*m+i]));
            }
        }

        //TODO: Think about synchronization


        // exchange halo data
        //TODO: Exchange halo data using the update directive




        // do one part on GPU       
        //TODO: adapt the present data on the GPU corresponding to the matrix decomposition
        //TODO: Enable asynchronous computations
        #pragma acc parallel loop present(Anew[0:n*m],A[0:n*m]) 
        for( j = 1; j < j_cpu_start; j++) {
            for( i = 1; i < m-1; i++ ) {
                A[j *m+ i] = Anew[j *m+ i];
            }
        }

        // do other part on CPU
        //TODO: To fully utilize the CPU use OpenMP to execute this loop in parallel

        for( j = j_cpu_start; j < n-1; j++) {
            for( i = 1; i < m-1; i++ ) {
                A[j *m+ i] = Anew[j *m+ i];
            }
        }


        //TODO: update defered the acc_err variable to the host

        //TODO: what to do with the two error values?


        if(iter % 10 == 0) {
            printf("%5d, %0.6f\n", iter, err);
        }

        iter++;
    } // end while
  } // end data region
    double runtime = omp_get_wtime() - starttime;


    printf(" total: %f s\n", runtime );

    free(y0);
    free(Anew);
    free(A);

    return 0;
}

#ifdef _OPENACC
static void initGPU(int argc, char** argv) {
        // gets the device id (if specified) to run on
        int devId = -1;
        if (argc > 1) {
                devId = atoi(argv[1]);
                int devCount = acc_get_num_devices(acc_device_nvidia);
                if (devId < 0 || devId >= devCount) {
                        printf("The specified device ID is not supported, the default ID is taken.\n");
			devId = -1;
                }
        }
        if (devId != -1) {
                acc_set_device_num(devId, acc_device_nvidia);
        }
        // creates a context on the GPU just to
        // exclude initialization time from computations
        acc_init(acc_device_nvidia);

        // print device id
        devId = acc_get_device_num(acc_device_nvidia);
        printf("Running on GPU with ID %d.\n", devId);

}
#endif
