/***********************************************
* Jacobi Solver with OpenACC
* ----------------------------------------------
* Authors: Sandra Wienke, RWTH Aachen University
* 
************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <omp.h>
#include <openacc.h>

static void initGPU(int argc, char** argv);

int main(int argc, char** argv)
{
    int i,j;
    int n = 4096;
    int m = 4096;
    int iter_max = 100;
    int n_cpu;
    if (argc > 2) n_cpu = atoi(argv[2]);
    else n_cpu = 4000;
    int j_cpu_start = n - n_cpu;
    if (j_cpu_start % 2 != 0 || n_cpu < 1 || n_cpu > n-2) {printf("Given number of rows on CPU is not valid.\n");exit(-1);}

    const double pi = 2.0 * asin(1.0);
    const double tol = 1.0e-5;
    double err = 1.0;
    double acc_err0 = 1.0;
    double acc_err1 = 1.0;

    double* restrict const A	= malloc( n*m * sizeof(double) );
    double* restrict const Anew	= malloc( n*m * sizeof(double) );
    double* restrict const y0	= malloc( n   * sizeof(double) );

#ifdef _OPENACC
  initGPU(argc, argv);
#endif

    // ensure memory locality on cc-NUMA with first touch
    #pragma omp parallel for default(none) shared(A,Anew,m,n,j_cpu_start) private(i,j)
    for( j = j_cpu_start-1; j < n; j++)
    {
        for( i = 0; i < m; i++ )
        {
            Anew[j *m+ i] 	= 0.0;
            A[j *m+ i] 		= 0.0;
        }
    }

    // set boundary conditions
    #pragma omp parallel for default(none) shared(A,m,n) private(i)
    for (i = 0; i < m; i++)
    {
    	A[0	    *m+ i] = 0.0;
    	A[(n-1) *m+ i] = 0.0;
    }

    #pragma omp parallel for default(none) shared(A,Anew,y0,m,n) private(j)
    for (j = 0; j < n; j++)
    {
        y0[j] = sin(pi * j / (n-1));
        A[j *m+ 0] = y0[j];
        A[j *m+ (m-1)] = y0[j]*exp(-pi);
        Anew[j *m+ 0] = A[j *m+ 0];
        Anew[j *m+ (m-1)] = A[j *m+ (m-1)];
    }

  // print in parallel region to get number of threads
  #pragma omp parallel
  {
    #pragma omp master
    {
      printf("Jacobi relaxation Calculation: %d x %d mesh with %d threads. %d rows out of %d on CPU.\n", n, m,omp_get_num_threads(),n_cpu,n);
    }
  }
    double starttime = omp_get_wtime();
    int iter = 0;

    // asynchronous data movement and kernel execution should be considered
    // However, PGI's current (14.2) compiler does not yet support async/wait clauses for all needed data movements

    //TODO: delete the data region and use corresponding unstructured data constructs
  #pragma acc data copy(A[0:(j_cpu_start+1)*m]) create(Anew[0:(j_cpu_start+1)*m])
  {
    //TODO: Move/create needed data on GPU 0: use unstructured data management
    //TODO: also create acc_err0 explicitly on GPU 0

    //TODO: Move/create needed data on GPU 1: use unstructured data management
    //TODO: also create acc_err1 explicitly on GPU 1

    while ( err > tol && iter < iter_max ) {

        acc_err0 = 0.0;
        acc_err1 = 0.0;
        err = 0.0;

        //TODO: update explicitly acc_err0 on GPU 0

        #pragma acc update device(acc_err0)

        //TODO: update explicitly acc_err1 on GPU 1

        #pragma acc update device(acc_err1)

        // do first part on GPU 0
        //TODO: Set GPU 0 and adapt the data range present on GPU 0

        #pragma acc parallel loop present(acc_err0,Anew[0:(j_cpu_start+1)*m],A[0:(j_cpu_start+1)*m]) async reduction(max:acc_err0)
        for( j = 1; j < j_cpu_start/2; j++) {
            for( i = 1; i < m-1; i++ ) {
                Anew[j *m+ i] = 0.25 * ( A[j     *m+ (i+1)] + A[j     *m+ (i-1)]
                                     +   A[(j-1) *m+ i]     + A[(j+1) *m+ i]);
                acc_err0 = fmax(acc_err0,fabs(Anew[j*m+i]-A[j*m+i]));
            }
        }

        // do second part on GPU 1
        //TODO: Set GPU 1 and compute corresponding data on GPU 1, set acc_err1 as present

        for( j = j_cpu_start/2; j < j_cpu_start; j++) {
            for( i = 1; i < m-1; i++ ) {
                Anew[j *m+ i] = 0.25 * ( A[j     *m+ (i+1)] + A[j     *m+ (i-1)]
                                     +   A[(j-1) *m+ i]     + A[(j+1) *m+ i]);
                acc_err1 = fmax(acc_err1,fabs(Anew[j*m+i]-A[j*m+i]));
            }
        }

        // do third part on CPU
        #pragma omp parallel for default(none) shared(A,Anew,m,n,j_cpu_start) private(i,j) reduction(max:err)
        for( j = j_cpu_start; j<n-1; j++) {
            for( i = 1; i < m-1; i++ ) {
                Anew[j *m+ i] = 0.25 * ( A[j     *m+ (i+1)] + A[j     *m+ (i-1)]
                                     +   A[(j-1) *m+ i]     + A[(j+1) *m+ i]);
                err = fmax(err,fabs(Anew[j*m+i]-A[j*m+i]));
            }
        }

        #pragma acc wait

        // exchange halo data
        //TODO: GPU 0 must update 1 line, GPU 1 must update 2 lines, CPU must update 1 line
        #pragma acc update host(  Anew[(j_cpu_start-1)*m +1:(m-2)])
        #pragma acc update device(Anew[ j_cpu_start   *m +1:(m-2)])


        // do first part on GPU 0    
        //TODO: Set GPU 0 and adapt the data range present on GPU 0

        #pragma acc parallel loop present(Anew[0:(j_cpu_start+1)*m],A[0:(j_cpu_start+1)*m]) async 
        for( j = 1; j < j_cpu_start/2; j++) {
            for( i = 1; i < m-1; i++ ) {
                A[j *m+ i] = Anew[j *m+ i];
            }
        }

        // do second part on GPU 1
        //TODO: Set GPU 1 and compute corresponding data on GPU 1

        for( j = j_cpu_start/2; j < j_cpu_start; j++) {
            for( i = 1; i < m-1; i++ ) {
                A[j *m+ i] = Anew[j *m+ i];
            }
        }


        // do third part on CPU
        #pragma omp parallel for default(none) shared(A,Anew,m,n,j_cpu_start) private(i,j)
        for( j = j_cpu_start; j < n-1; j++) {
            for( i = 1; i < m-1; i++ ) {
                A[j *m+ i] = Anew[j *m+ i];
            }
        }

        #pragma acc wait

        //TODO: update explicitly acc_err0 from GPU 0 to the host

        #pragma acc update host(acc_err0)

        //TODO: update explicitly acc_err1 from GPU 1 to the host

        #pragma acc update host(acc_err1)

        // combine both error values
        //TODO: Combine all three error values to one value
        err = fmax(err,acc_err0);

        if(iter % 10 == 0) {
            printf("%5d, %0.6f\n", iter, err);
        }

        iter++;
    } // end while
  } // end data region

    //TODO: Delete/copyout needed data from GPU 0: use unstructured data management

    //TODO: Delete/copyout needed data from GPU 1: use unstructured data management


    double runtime = omp_get_wtime() - starttime;


    printf(" total: %f s\n", runtime );


    free(y0);
    free(Anew);
    free(A);

    return 0;
}

#ifdef _OPENACC
static void initGPU(int argc, char** argv) {
        // gets the device id (if specified) to run on
        int devId = -1;
        int devCount = acc_get_num_devices(acc_device_nvidia);
        if (argc > 1) {
                devId = atoi(argv[1]);
                if (devId < 0 || devId >= devCount) {
                        printf("The specified device ID is not supported, the default ID is taken.\n");
			devId = -1;
                }
        }
	if (devId == -1) devId = 0;

	// creates a context on both GPUs
	// excludes initialization time from computations
        acc_set_device_num(devId, acc_device_nvidia);
	acc_init(acc_device_nvidia);
	acc_set_device_num((devId+1)%devCount, acc_device_nvidia); 
	acc_init(acc_device_nvidia);
        acc_set_device_num(devId, acc_device_nvidia);

        // print device id
        devId = acc_get_device_num(acc_device_nvidia);
        printf("Running on GPUs with IDs %d and %d.\n", devId,(devId+1)%devCount);

}
#endif
